#
# remt - note-taking support tool for reMarkable tablet
#
# Copyright (C) 2018-2021 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Drawing tool unit tests.
"""

import operator

from remt import tool
from remt.data import ColorIndex, Pen, Segment, Stroke

def test_group_line_segments() -> None:
    """
    Test grouping line segments.
    """
    items = [
        (0, 0, 1),
        (1, 1, 1),
        (2, 2, 2),
        (3, 3, 3),
        (4, 4, 3),
        (5, 5, 3),
        (6, 6, 4),
        (7, 7, 4),
        (8, 8, 4),
    ]
    s1, s2, s3, s4 = tool.group_line_segments(iter(items), operator.itemgetter(2))
    assert (1, [(0, 0), (1, 1), (2, 2)]) == s1
    assert (2, [(2, 2), (3, 3)]) == s2
    assert (3, [(3, 3), (4, 4), (5, 5), (6, 6)]) == s3
    assert (4, [(6, 6), (7, 7), (8, 8)]) == s4

def test_group_line_segments_end_point() -> None:
    """
    Test grouping line segments with single point at the end.
    """
    items = [
        (0, 0, 1),
        (1, 1, 1),
        (2, 2, 2),
        (3, 3, 3),
        (4, 4, 3),
        (5, 5, 3),
        (6, 6, 0),
    ]
    s1, s2, s3 = tool.group_line_segments(iter(items), operator.itemgetter(2))
    assert (1, [(0, 0), (1, 1), (2, 2)]) == s1
    assert (2, [(2, 2), (3, 3)]) == s2
    assert (3, [(3, 3), (4, 4), (5, 5), (6, 6)]) == s3

def test_group_line_segments_empty() -> None:
    """
    Test grouping line segments with empty input.
    """
    items = tool.group_line_segments(iter([]), operator.itemgetter(2))
    assert next(items, None) is None

def test_line_const() -> None:
    """
    Test calculation line with constant width and pressure.
    """
    stroke = Stroke(
        0, Pen(0), ColorIndex(0), 10,
        [
            create_segment(0, 0, 100, 0),
            create_segment(1, 1, 100, 0),
            create_segment(2, 2, 100, 0),
        ]
    )
    calc = lambda v: v.width * 2
    items = tool.line_const(calc, stroke)
    result = next(items)

    assert next(items, None) is None

    assert 20 == result.width
    assert 1 == result.pressure
    assert [(0, 0), (1, 1), (2, 2)] == result.points

def test_line_var_pressure() -> None:
    """
    Test calculation of line with constant width and varying pressure.
    """
    stroke = Stroke(
        0, Pen(0), ColorIndex(0), 10,
        [
            create_segment(0, 0, 100, 0.1),
            create_segment(1, 1, 100, 0.2),
            create_segment(2, 2, 100, 0.2),
            create_segment(3, 3, 100, 0.2),
            create_segment(4, 4, 100, 0.3),
            create_segment(5, 5, 100, 0),
        ]
    )
    calc = lambda st: st.width * 2
    l1, l2, l3 = tool.line_var_pressure(calc, stroke)

    assert 20 == l1.width
    assert 20 == l2.width
    assert 20 == l3.width

    assert 10 == l1.pressure
    assert 20 == l2.pressure
    assert 30 == l3.pressure

    assert [(0, 0), (1, 1)] == l1.points
    assert [(1, 1), (2, 2), (3, 3), (4, 4)] == l2.points
    assert [(4, 4), (5, 5)] == l3.points

def create_segment(
        x: float, y: float, width: float, pressure: float
) -> Segment:
    return Segment(0, x, y, 0, 0, width, pressure)

# vim: sw=4:et:ai



#
# remt - note-taking support tool for reMarkable tablet
#
# Copyright (C) 2018-2021 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
reMarkable tablet lines format parser.
"""

import contextvars
import cytoolz.itertoolz as itz  # type: ignore
import enum
import json
import logging
import struct
import typing as tp

from . import data as rdata
from .util import flatten

logger = logging.getLogger(__name__)

T = tp.TypeVar('T')
HEADER_START = b'reMarkable .lines file, version='

CTX_PARSER = contextvars.ContextVar['Version']('version')

TYPE_HEADER_PAGE = tp.Tuple[bytes, int, bytes]
TYPE_PAGE = tp.Tuple[int, int, int]
TYPE_LAYER = tp.Tuple[int]
TYPE_STROKE_V5 = tp.Tuple[int, int, int, float, int, int]
TYPE_STROKE = tp.Tuple[int, int, int, float, int]
TYPE_SEGMENT = tp.Tuple[float, float, float, float, float, float]

@enum.unique
class Version(enum.IntEnum):
    """
    Version of format of reMarkable file.
    """
    VERSION_3 = ord('3')
    VERSION_5 = ord('5')

class Parser(tp.Generic[T]):
    """
    Parser of reMarkable file binary data.

    :var parser: Binary data Struct object.
    """
    def __init__(self, spec: str):
        self.parser = struct.Struct(spec)

    def parse(self, fin: tp.BinaryIO) -> T:
        """
        Read number of bytes from a file and parse the data of a reMarkable
        document item using binary data Struct object.

        :param fin: File object.
        """
        buff = fin.read(self.parser.size)
        return tp.cast(T, self.parser.unpack(buff))

# define parsers for document items found in reMarkable table file format
FMT_HEADER_PAGE = Parser[TYPE_HEADER_PAGE](
    '<{}sB10s'.format(len(HEADER_START))
)
FMT_PAGE = Parser[TYPE_PAGE]('<BBH')  # TODO: might be 'I'
FMT_LAYER = Parser[TYPE_LAYER]('<I')
FMT_STROKE = Parser[TYPE_STROKE]('<IIIfI')
FMT_STROKE_V5 = Parser[TYPE_STROKE_V5]('<IIIfII')
FMT_SEGMENT = Parser[TYPE_SEGMENT]('<ffffff')

def parse(
        data: tp.BinaryIO,
        page_number: int,
        highlights: tp.List[rdata.Stroke]
    ) -> rdata.Items:
    """
    Parse reMarkable file and return iterator of document items.

    :param data: reMarkable file data.
    :param page_number: Page number for which data is parsed.
    :param highlights: Highlighter information parsed from JSON files.
    """
    header, ver, *_ = FMT_HEADER_PAGE.parse(data)
    assert header.startswith(HEADER_START), 'header is {!r}'.format(header)
    CTX_PARSER.set(Version(ver))
    logger.info('file format version {}'.format(chr(ver)))
    yield from parse_page(data, page_number, highlights)

def empty_page(page_number: int) \
        -> tp.Iterable[tp.Union[rdata.Page, rdata.PageEnd]]:
    """
    Generate empty page for document rendering.

    :param page_number: Page number to be associated with the page.
    """
    yield rdata.Page(page_number)
    yield rdata.PageEnd(page_number)

def parse_page(
        data: tp.BinaryIO,
        page_number: int,
        highlights: tp.List[rdata.Stroke]
    ) -> rdata.Items:
    """
    Parse page from reMarkable table file.
    """
    n, _, _ = FMT_PAGE.parse(data)
    items = (parse_layer(i, data) for i in range(n))

    yield rdata.Page(page_number)
    yield from flatten(items)
    yield from highlights
    yield rdata.PageEnd(page_number)

def parse_layer(n_layer: int, data: tp.BinaryIO) -> rdata.Items:
    n, = FMT_LAYER.parse(data)

    items = (parse_stroke(i, data) for i in range(n))

    yield rdata.Layer(n_layer)
    yield from flatten(items)

def parse_stroke(n_stroke: int, data: tp.BinaryIO) \
        -> tp.Iterable[rdata.Stroke]:

    ver = CTX_PARSER.get()
    if ver == Version.VERSION_5:
        pen, color_idx, unk1, width, unk2, n = FMT_STROKE_V5.parse(data)
    elif ver == Version.VERSION_3:
        pen, color_idx, a, width, n = FMT_STROKE.parse(data)
    else:
        assert False

    # fix color index for highlighter, default it to yellow if black, gray
    # or white was stored by remarkable
    if rdata.Pen.is_highlighter(pen) and color_idx in (0, 1, 2):
        color = rdata.ColorIndex.YELLOW
    else:
        color = rdata.ColorIndex(color_idx)

    segments = [parse_segment(i, data) for i in range(n)]
    stroke = rdata.Stroke(
        n_stroke,
        rdata.Pen(pen),
        color,
        width,
        segments
    )

    yield stroke

def parse_segment(n_seg: int, data: tp.BinaryIO) -> rdata.Segment:
    x, y, speed, direction, width, pressure = FMT_SEGMENT.parse(data)
    return rdata.Segment(n_seg, x, y, speed, direction, width, pressure)

def parse_highlights(fin: tp.TextIO) -> tp.Iterator[rdata.Stroke]:
    data = json.load(fin)
    highlights = itz.concat(v for v in data.get('highlights', []))
    items = (parse_highlight_item(h) for h in highlights)
    yield from items

def parse_highlight_item(highlight: dict[str, tp.Any]) -> rdata.Stroke:
    to_seg = lambda x, y, w: rdata.Segment(0, x, y, 0, 0, w, 1)
    items = (
        (v['x'], v['y'], v['width'], v['height'] / 2)
        for v in highlight['rects']
    )
    segments = itz.concat(
        (to_seg(x, y + h, h), to_seg(x + w, y + h, h))  # type: ignore
        for x, y, w, h in items
    )
    default_color = rdata.ColorIndex.YELLOW
    color = highlight.get('color', default_color)
    # sometimes remarkable might save highlighter color as black, revert to
    # default then
    color = default_color if color == rdata.ColorIndex.BLACK else color
    return rdata.Stroke(
        0,
        rdata.Pen.HIGHLIGHTER_2,
        color,
        rdata.PEN_WIDTH_HIGHLIGHTER,
        list(segments),
        text=highlight['text']
    )

# vim: sw=4:et:ai



#
# remt - note-taking support tool for reMarkable tablet
#
# Copyright (C) 2018-2021 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
PDF utilities.
"""

import gi  # type: ignore
gi.require_version('Poppler', '0.18')  # noqa: E402

import pathlib  # noqa: E402
from gi.repository import Poppler  # type: ignore  # noqa: E402

from . import const  # noqa: E402
from .data import TextSelection, Stroke  # noqa: E402


def pdf_open(fn: str) -> Poppler.Document:
    """
    Open PDF file and return Poppler library PDF document.

    :param fn: PDF file name.
    """
    path = pathlib.Path(fn).resolve().as_uri()
    return Poppler.Document.new_from_file(path)

def pdf_scale(page: Poppler.Page) -> float:
    """
    Get scaling factor for a PDF page to fit reMarkable tablet vector data
    onto the page.

    :param page: Poppler PDF page object.
    """
    w, h = page.get_size()
    return max(w / const.PAGE_WIDTH, h / const.PAGE_HEIGHT)  # type: ignore

def pdf_area(page: Poppler.Page, selection: TextSelection) \
        -> Poppler.Rectangle:
    """
    Get PDF page area for a text selection data.

    :param page: Poppler PDF page object.
    :param selection: Text selection data.
    """
    factor = pdf_scale(page)
    y = selection.y * factor

    area = Poppler.Rectangle()
    area.x1 = selection.x1 * factor
    area.y1 = y - 1
    area.x2 = selection.x2 * factor
    area.y2 = y + 1

    return area

def pdf_text(page: Poppler.Page, selection: TextSelection) -> str:
    """
    Having a reMarkable tablet stroke data, get text annotated by the
    stroke.

    :param page: Poppler PDF page object.
    :param selection: Text selection data.
    """
    area = pdf_area(page, selection)
    return page.get_text_for_area(area)  # type: ignore

def pdf_highlight_text(
        doc: Poppler.Document, page: int, stroke: Stroke
    ) -> None:
    """
    Create highlight text annotation in PDF ddocument.

    :param doc: Poppler PDF document object.
    :param page: Page index.
    :param stroke: reMarkable tablet stroke information.
    """
    pdf_page = doc.get_page(page)
    factor = pdf_scale(pdf_page)
    _, height = pdf_page.get_size()

    p1 = stroke.segments[0]
    p2 = stroke.segments[0]
    area = Poppler.Rectangle()
    y = p1.y * factor
    area.x1 = p1.x * factor
    area.y1 = y - 1
    area.x2 = p2.x * factor
    area.y2 = y + 1

    region = pdf_page.get_selected_region(
        1.0, Poppler.SelectionStyle.GLYPH, area
    )
    quads = []
    for i in range(region.num_rectangles()):
        r = region.get_rectangle(i)
        q = Poppler.Quadrilateral()

        x1, y1 = r.x + r.width, height - r.y - r.height
        x2, y2 = r.x, height - r.y
        q.p1 = poppler_point(x1, y1)
        q.p2 = poppler_point(x2, y1)
        q.p3 = poppler_point(x1, y2)
        q.p4 = poppler_point(x2, y2)
        quads.append(q)

    if quads:
        extents = region.get_extents()
        bbox = Poppler.Rectangle()
        bbox.x1 = extents.x
        bbox.x2 = extents.x + extents.width
        bbox.y1 = height - extents.y
        bbox.y2 = height - extents.y - extents.height

        markup = Poppler.AnnotTextMarkup.new_highlight(doc, bbox, quads)
        color = Poppler.Color()
        color.red = 65500
        color.green = 0
        color.blue = 0
        markup.set_color(color)
        markup.set_label(stroke.text)
        markup.set_opacity(0.2)
        pdf_page.add_annot(markup)

def poppler_point(x: float, y: float) -> Poppler.Point:
    """
    Convert coordinates to Poppler point object.
    """
    p = Poppler.Point()
    p.x = x
    p.y = y
    return p

# vim: sw=4:et:ai



#
# remt - note-taking support tool for reMarkable tablet
#
# Copyright (C) 2018-2021 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
PDF annotations indexer.
"""

import cytoolz.functoolz as ftz  # type: ignore
import textwrap
import typing as tp
from functools import partial, reduce
from statistics import fmean

from .data import DocumentItem, Items, Pen, Page, Stroke, TextSelection, \
    PEN_WIDTH_HIGHLIGHTER
from .pdf import Poppler, pdf_text
from .util import split, to_point_y

FMT_PAGE = 'Page {} ({})'.format

PageStrokes = tp.Iterable[tp.Tuple[Page, tp.Tuple[Stroke]]]
PageTexts = tp.Iterable[tp.Tuple[Poppler.Page, tp.Iterable[str]]]

def ann_text(pdf_doc: Poppler.Document, items: Items) -> PageTexts:
    """
    Get annotated text from PDF document using document items parsed from
    reMarkable file.
    """
    get_page = pdf_doc.get_page
    strokes = page_strokes(items)
    selections = ((p, page_selections(s)) for p, s in strokes)

    # get PDF pages
    pages = ((get_page(p.number), s) for p, s in selections)

    # for each page and stroke get text under stroke
    texts = ((p, t) for p, s in pages if (t := page_text(p, s)))
    yield from texts

def page_strokes(items: Items) -> PageStrokes:
    """
    Get reMarkable tablet page and stroke information.
    """
    is_page = ftz.flip(isinstance, Page)

    # find pages and strokes
    items = (v for v in items if is_page(v) or is_highlighter(v))

    # split into (page, strokes)
    yield from split(is_page, items)  # type: ignore

def page_selections(strokes: tp.Tuple[Stroke]) -> tp.List[TextSelection]:
    """
    Detect line selections from reMarkable table stroke data.

    The line selections are sorted and overlapping line selections are
    merged.
    """
    selections = [sel for st in strokes if (sel := to_line_selection(st))]
    selections = sorted(selections)
    return merge_line_selection(selections)

def page_text(
        page: Poppler.Page, selections: tp.List[TextSelection]
) -> tp.List[str]:
    """
    Get text for each page stroke.

    Empty strings are filtered out.

    :param page: Poppler PDF page object.
    :param selections: Collection of text selections.
    """
    to_text = partial(pdf_text, page)
    return [t for s in selections if (t := to_text(s))]

def fmt_ann_text(items: PageTexts) -> tp.Iterable[tp.Tuple[str, str]]:
    """
    Format annotated text read from PDF document as reStructuredText
    document.
    """
    texts = ((fmt_header(p), fmt_text(t)) for p, t in items)
    yield from texts

def fmt_header(page: Poppler.Page) -> str:
    """
    Format page header in reStructuredText format.
    """
    header = FMT_PAGE(page.get_label(), page.get_index())
    return '{}\n{}'.format(header, '=' * len(header))

def fmt_text(texts: tp.Iterable[str]) -> str:
    """
    Format collection of text items in reStructuredText format.
    """
    result = '\n'.join(texts)
    result = textwrap.indent(result, ' ' * 4)
    result = '::\n\n{}\n'.format(result)
    return result

def is_highlighter(item: DocumentItem) -> bool:
    """
    Check if reMarkable tablet data item is a highlighter.

    :pram item: reMarkable data item data.
    """
    return isinstance(item, Stroke) and Pen.is_highlighter(item.pen)

def is_horizontal_line(
        y_min: float,
        y_max: float,
        ys: float,
        width: float,
    ) -> bool:
    """
    Check if stroke with y-axis coordinates between `y_min` and `y_max`
    values can be treated as horizontal line at `ys` of certain width.

    :param y_min: Minimum y-axis coordinate of a candidate line.
    :param y_max: Maximum y-axis coordinate of a candidate line.
    :param ys: Horizontal line y-axis coordinate.
    :param width: Horizontal line width.
    """
    w = width / 2
    return y_min >= ys - w and ys + w >= y_max

def to_line_selection(stroke: Stroke) -> tp.Optional[TextSelection]:
    """
    Create text selection data from a reMarkable tablet stroke information.

    If stroke does not create a horizontal line, then null is returned.

    :param stroke: reMarkable tablet stroke information.
    """
    x1 = min(s.x for s in stroke.segments)
    x2 = max(s.x for s in stroke.segments)
    y1 = min(s.y for s in stroke.segments)
    y2 = max(s.y for s in stroke.segments)

    ys = fmean(to_point_y(s) for s in stroke.segments)

    w = 1.1 * PEN_WIDTH_HIGHLIGHTER / 2  # allow 10% wiggle room
    if is_horizontal_line(y1, y2, ys, w):
        return TextSelection(x1, x2, ys)
    else:
        return None

def merge_line_selection(selections: tp.Iterable[TextSelection]) \
        -> tp.List[TextSelection]:
    """
    Merge text selection data if text selection data overlaps.

    The input should be sorted using text selection ordering.
    """
    return reduce(merge_line_selection_item, selections, [])

def merge_line_selection_item(acc: tp.List[TextSelection], ls: TextSelection) \
        -> tp.List[TextSelection]:

    to_merge = acc and acc[-1].overlaps(ls, PEN_WIDTH_HIGHLIGHTER)
    result, item = (acc[:-1], acc[-1].merge(ls)) if to_merge else (acc, ls)
    return result + [item]

# vim: sw=4:et:ai



#
# remt - note-taking support tool for reMarkable tablet
#
# Copyright (C) 2018-2021 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Command line commands.
"""

import argparse
import asyncssh
import configparser
import json
import logging
import os.path
import shutil
import typing as tp
import urllib.request
from contextlib import asynccontextmanager
from collections import namedtuple
from cytoolz.dicttoolz import get_in  # type: ignore
from datetime import datetime
from iterfzf import iterfzf
from functools import partial
from uuid import uuid4 as uuid

import remt
from .data import Items, _ERRORS
from .error import ConfigError, FileError, ConnectionError
from .indexer import ann_text, fmt_ann_text
from .meta import Meta, MetaDir, BASE_DIR, fn_path, cache_dir, read_meta, \
    sync_meta, fn_metadata, create_metadata
from .parser import parse_highlights
from .pdf import pdf_open
from .util import flatten

logger = logging.getLogger(__name__)

FILE_TYPE = {
    'CollectionType': 'd',
}

ERROR_CFG = """\
Configuration file {conf} not found

  Create configuration file `{conf}` with contents

    [connection]
    host=10.11.99.1
    user=root
    password=<your reMarkable tablet password>
"""


# config: remt config
# sftp: SFTP connection to a device
# dir_meta: metadata files
# meta: parsed metadata
# dir_data: directory where to fetch files from a device or where to
#   prepare files for upload
RemtContext = namedtuple(
    'RemtContext',
    ['config', 'sftp', 'dir_meta', 'meta', 'dir_data'],
)

#
# utilities
#

def read_config() -> configparser.ConfigParser:
    """
    Read and return Remt project configuration.
    """
    conf_file = os.path.expanduser('~/.config/remt.ini')
    if not os.path.exists(conf_file):
        msg = ERROR_CFG.format(conf=conf_file)
        raise ConfigError(msg)

    cp = configparser.ConfigParser()
    cp.read(conf_file)
    return cp

@asynccontextmanager
async def remt_ctx(
        dir_base: tp.Optional[str]=None
    ) -> tp.AsyncIterator[RemtContext]:
    """
    Create a Remt project context.

    The function is an asynchronous context manager.
    """
    config = read_config()
    host = config.get('connection', 'host')
    user = config.get('connection', 'user')
    password = config.get('connection', 'password')

    if dir_base is None:
        dir_base = cache_dir(host)

    mkdir = partial(os.makedirs, exist_ok=True)
    dir_meta = os.path.join(dir_base, 'metadata')
    dir_data = os.path.join(dir_base, 'data')

    mkdir(dir_meta)
    mkdir(dir_data)

    try:
        ctx_conn = asyncssh.connect(host, username=user, password=password)
        async with ctx_conn as conn:
            async with conn.start_sftp_client() as sftp:
                await sync_meta(sftp, dir_meta)
                meta = read_meta(dir_meta)
                yield RemtContext(config, sftp, dir_meta, meta, dir_data)
    except OSError as ex:
        if ex.errno == 101:
            raise ConnectionError(
                'Cannot connect to a reMarkable tablet: {}'.format(ex.strerror)
            )
        else:
            raise
    finally:
        shutil.rmtree(dir_data)
        for msg in _ERRORS:
            logger.warning(msg)

def norm_path(path: str) -> str:
    """
    Normalise path of a file on a reMarkable tablet.

    All leading and trailing slashes are removed. Multiple slashes are
    replaced with one.

    :param path: Path to normalise.
    """
    return os.path.normpath(path).strip('/')

def fn_interactive(ctx: RemtContext, is_dir: bool=False) -> str:
    """
    Get filename interactively using `fzf` program.

    Raise error if selection interrupted.

    :param ctx: Remt project context.
    :param is_dir: Allow to select only directories if true.
    """
    if is_dir:
        files = (
            k for k, v in ctx.meta.items()
            if v['type'] == 'CollectionType'
        )
    else:
        files = ctx.meta.keys()

    result = iterfzf(sorted(files))  # type: ignore
    if result is None:
        raise FileError('Interactive file selection cancelled')

    return result  # type: ignore

#
# parsing pages from a collection of files in reMarkable lines format
#

def parse_document(ctx: RemtContext, data: Meta) -> Items:
    get_page_fn = partial(get_page_filename, ctx.dir_data, data['uuid'], 'rm')
    get_highlights_fn = partial(
        get_page_filename,
        ctx.dir_data,
        '{}.highlights'.format(data['uuid']),
        'json'
    )
    pages = data['content'].get('pages')
    if pages is None:
        pages = [str(i) for i in range(data['content']['pageCount'])]

    items = flatten(  # type: ignore
        parse_page(get_page_fn(p), get_highlights_fn(p), i)
        for i, p in enumerate(pages)  # type: ignore
    )
    yield from items


def parse_page(fin: str, fin_highlights: str, page_number: int) -> Items:
    """
    Parse page from reMarkable lines file.

    Return empty page if file does not exist.

    :param fin: reMarkable lines file.
    :param fin_highlights: reMarkable highlights file.
    :param page_number: Page number to be associated with the page.
    """
    if os.path.exists(fin):
        # load highlights information for the page
        if os.path.exists(fin_highlights):
            with open(fin_highlights, 'r') as ft:
                highlights = list(parse_highlights(ft))
        else:
            highlights = []

        with open(fin, 'rb') as fb:
            yield from remt.parse(fb, page_number, highlights)
    else:
        yield from remt.empty_page(page_number)

def get_page_filename(
        base_dir: str,
        uuid: str,
        ext: str,
        page_uuid: str
    ) -> str:
    path = os.path.join(base_dir, uuid, page_uuid)
    return '{}.{}'.format(path, ext)

#
# cmd: ls
#

def marker(cond: bool, marker: str) -> str:
    return marker if cond else '-'

def ls_line(fn: str, data: Meta) -> str:
    """
    Create `ls` command basic output line.
    """
    return fn

def ls_line_long(fn: str, data: Meta) -> str:
    """
    Create `ls` command long output line.
    """
    bookmarked = marker(data['pinned'] is True, 'b')
    is_dir = marker(data['type'] == 'CollectionType', 'd')

    tstamp = int(data['lastModified']) / 1000
    tstamp_ts = datetime.fromtimestamp(tstamp)

    line = '{}{} {:%Y-%m-%d %H:%M:%S} {}'.format(
        is_dir, bookmarked, tstamp_ts, fn
    )
    return line

def ls_filter_path(meta: MetaDir, path: str) -> MetaDir:
    """
    Filter metadata to keep metadata starting with path name.

    The path name itself *is* filtered out.

    :param meta: reMarkable tablet metadata.
    :param path: Path name.
    """
    meta = {k: v for k, v in meta.items() if k.startswith(path) and path != k}
    return meta

def ls_filter_parent_uuid(meta: MetaDir, uuid: tp.Optional[str]) -> MetaDir:
    """
    Filter metadata to keep metadata starting, which parent is identified
    by UUID.

    :param meta: reMarkable tablet metadata.
    :param uuid: UUID of parent directory.
    """
    check = lambda v: not v if uuid is None else v == uuid
    meta = {
        k: v for k, v in meta.items() if check(v.get('parent'))  # type: ignore
    }
    return meta

async def cmd_ls(args: argparse.Namespace) -> None:
    to_line = ls_line_long if args.long else ls_line
    path = norm_path(args.path) if args.path else None

    async with remt_ctx() as ctx:
        meta = ctx.meta

        if path == '%i':
            path = fn_interactive(ctx)

        # get starting UUID while we have all metadata
        start = None
        if path:
            start = fn_metadata(meta, path)['uuid']
            meta = ls_filter_path(meta, path)

        if not args.recursive:
            meta = ls_filter_parent_uuid(meta, start)

        lines = (to_line(k, v) for k, v in sorted(meta.items()))
        print('\n'.join(lines))

#
# cmd: mkdir
#

async def cmd_mkdir(args: argparse.Namespace) -> None:
    """
    Create a directory on reMarkable tablet device.
    """
    async with remt_ctx() as ctx:
        meta = ctx.meta
        path = norm_path(args.path)

        if path in meta:
            msg = 'Cannot create directory "{}" as it exists'.format(path)
            raise FileError(msg)

        parent, name = os.path.split(path)
        if parent and parent not in meta:
            raise FileError('Parent directory not found')

        assert bool(name)

        parent_uuid = get_in([parent, 'uuid'], meta)
        data = create_metadata(True, parent_uuid, name)

        dir_fn = os.path.join(ctx.dir_data, str(uuid()))

        with open(dir_fn + '.metadata', 'w') as f:
            json.dump(data, f)

        # empty content file required to create a directory
        with open(dir_fn + '.content', 'w') as f:
            json.dump({}, f)

        await ctx.sftp.mput(dir_fn + '.*', BASE_DIR)

#
# cmd: export
#
async def cmd_export(args: argparse.Namespace) -> None:
    path = norm_path(args.input)
    f_export = _export_remt if args.renderer == 'remt' else _export_rm

    async with remt_ctx() as ctx:
        if path == '%i':
            path = fn_interactive(ctx)

        output = args.output
        if os.path.isdir(output):
            output = os.path.join(output, os.path.basename(path))
            if not output.endswith('.pdf'):
                output += '.pdf'

        data = fn_metadata(ctx.meta, path)
        await f_export(ctx, data, output)

async def _export_remt(ctx: RemtContext, data: Meta, fout: str) -> None:
    """
    Export notebook or PDF document using Remt renderer.

    :param ctx: Remt project context.
    :param data: Metadata of input file.
    :param fout: Filename of output file.
    """
    to_copy = fn_path(data, ext='*')
    await ctx.sftp.mget(to_copy, ctx.dir_data, recurse=True)

    fin_pdf = fn_path(data, base=ctx.dir_data, ext='.pdf')
    pdf_doc = pdf_open(fin_pdf) if os.path.exists(fin_pdf) else None

    items = parse_document(ctx, data)
    with remt.draw_context(pdf_doc, fout) as ctx:  # type: ignore
        for item in items:
            remt.draw(item, ctx)  # type: ignore

async def _export_rm(ctx: RemtContext, data: Meta, fout: str) -> None:
    """
    Export notebook or PDF document using reMarkable tablet device.

    :param ctx: Remt project context.
    :param data: Metadata of input file.
    :param fout: Filename of output file.
    """
    host = ctx.config.get('connection', 'host')
    uuid = data['uuid']
    url = 'http://{}/download/{}/placeholder'.format(host, uuid)

    response = urllib.request.urlopen(url)
    with open(fout, 'wb') as f:
        read = lambda: response.read(1024 ** 2)
        for f_data in iter(read, b''):
            f.write(f_data)

#
# cmd: import
#
def _prepare_import_data(ctx: RemtContext, fn_in: str, out_uuid: str) -> str:
    """
    Prepare import data for a file to be uploaded onto a reMarkable
    tablet.

    :param ctx: Remt project context.
    :param fn_in: File to be imported.
    :param out_uuid: UUID of the output directory located on a reMarkable
        tablet.
    """
    name = os.path.basename(fn_in)
    fn_base = os.path.join(ctx.dir_data, str(uuid()))
    data = create_metadata(False, out_uuid, name)

    fn_pdf = fn_base + '.pdf'
    shutil.copy(fn_in, fn_pdf)
    with open(fn_base + '.metadata', 'w') as f:
        json.dump(data, f)

    # empty content file required
    with open(fn_base + '.content', 'w') as f:
        page_count = pdf_open(fn_pdf).get_n_pages()
        content = {
            'fileType': 'pdf',
            'lastOpenedPage': 0,
            'lineHeight': -1,
            'pageCount': page_count,

        }
        json.dump(content, f)
    return fn_base + '.*'

async def cmd_import(args: argparse.Namespace) -> None:
    """
    Import a number of files onto a directory on a reMarkable tablet.
    """
    output = norm_path(args.output)

    async with remt_ctx() as ctx:
        if output == '%i':
            output = fn_interactive(ctx, is_dir=True)

        out_meta = fn_metadata(ctx.meta, output)
        if out_meta['type'] != 'CollectionType':
            raise FileError('Destination path is not a directory')

        out_uuid = out_meta['uuid']
        to_import = [
            _prepare_import_data(ctx, fn, out_uuid) for fn in args.input
        ]
        await ctx.sftp.mput(to_import, BASE_DIR)

#
# cmd: index
#

async def cmd_index(args: argparse.Namespace) -> None:
    path = norm_path(args.input)

    async with remt_ctx() as ctx:
        if path == '%i':
            path = fn_interactive(ctx)

        data = fn_metadata(ctx.meta, path)

        to_copy = fn_path(data, ext='*')
        await ctx.sftp.mget(to_copy, ctx.dir_data, recurse=True)

        fin_pdf = fn_path(data, base=ctx.dir_data, ext='.pdf')
        pdf_doc = pdf_open(fin_pdf)

        items = parse_document(ctx, data)
        items = ann_text(pdf_doc, items)  # type: ignore
        items = fmt_ann_text(items)  # type: ignore

        for header, texts in items:  # type: ignore
            print(header)
            print(texts)

COMMANDS: dict[str, tp.Callable[..., tp.Any]] = {
    'ls': cmd_ls,
    'mkdir': cmd_mkdir,
    'export': cmd_export,
    'import': cmd_import,
    'index': cmd_index,
}

# vim: sw=4:et:ai



#
# remt - note-taking support tool for reMarkable tablet
#
# Copyright (C) 2018-2021 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
PDF annotation indexer unit tests.
"""

from .. import indexer as ridx
from ..data import ColorIndex, Page, Pen, Segment, Stroke, TextSelection

from unittest import mock

def test_page_text() -> None:
    """
    Test getting text for each page stroke.
    """
    page = mock.MagicMock()
    with mock.patch.object(ridx, 'pdf_text') as mock_pdf_text:
        mock_pdf_text.side_effect = ['a', '', 'b', 'c']

        # note: number of strokes the same as number of above side effects
        ts = [TextSelection(i, i, 0) for i in range(1, 5)]
        result = ridx.page_text(page, ts)

        # empty strings are removed
        assert ['a', 'b', 'c'] == result

def test_fmt_header() -> None:
    """
    Test formatting page header in reStructuredText format.
    """
    page = mock.MagicMock()
    page.get_label.return_value = 'IX'
    page.get_index.return_value = '10'

    result = ridx.fmt_header(page)
    print(result)
    assert 'Page IX (10)\n============' == result

def test_fmt_text() -> None:
    """
    Test formatting a collection of texts in reStructuredText format.
    """
    # note: no need to test for empty strings as these are removed by
    # page_text function
    texts = ['abc', 'xyz\nzyx']
    result = ridx.fmt_text(texts)
    expected = """\
::

    abc
    xyz
    zyx
"""
    assert expected == result

def test_is_highlighter() -> None:
    """
    Check if an object is a reMarkable tablet stroke and is a highlighter.
    """
    assert not ridx.is_highlighter(Page(1))
    assert not ridx.is_highlighter(Stroke(1, Pen.FINELINER_1, ColorIndex(1), 2, []))
    assert ridx.is_highlighter(Stroke(1, Pen.HIGHLIGHTER_1, ColorIndex(1), 2, []))

def test_is_horizontal_line() -> None:
    """
    Test checking if a candidate is a horizontal line.
    """
    assert not ridx.is_horizontal_line(1, 10, 5, 3)
    assert ridx.is_horizontal_line(1, 10, 5, 10)

def test_to_line_selection() -> None:
    """
    Test converting reMarkable tablet stroke information to text selection
    data.
    """
    stroke = Stroke(
        1, Pen.HIGHLIGHTER_1, ColorIndex(1), 2,
        [
            create_segment(10, 10),
            create_segment(100, 15),
            create_segment(200, 20),
        ]
    )

    result = ridx.to_line_selection(stroke)

    assert result is not None
    assert 10 == result.x1
    assert 200 == result.x2
    assert 15 == result.y

def test_to_line_selection_none() -> None:
    """
    Test converting reMarkable tablet stroke information to null instead of
    text selection data.
    """
    stroke = Stroke(
        1, Pen.HIGHLIGHTER_1, ColorIndex(1), 2,
        [
            create_segment(10, 0),
            create_segment(100, 40),
            create_segment(200, 15),
        ]
    )
    result = ridx.to_line_selection(stroke)
    assert result is None

def create_segment(x: float, y: float) -> Segment:
    return Segment(1, x, y, 0, 0, 30, 1)

# vim: sw=4:et:ai



#
# remt - note-taking support tool for reMarkable tablet
#
# Copyright (C) 2018-2021 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Data classes representing reMarkable tablet information.
"""

from __future__ import annotations

import cairo
import dataclasses as dtc
import enum
import operator
import typing as tp
from functools import total_ordering

# pen width used by reMarkable tablet tools
PEN_WIDTH_UKNOWN = 1
PEN_WIDTH_ERASE_AREA = 1
PEN_WIDTH_HIGHLIGHTER = 30

TEXT_SELECTION_ORDER = operator.attrgetter('y', 'x1', 'x2')

BrushPatterns: tp.TypeAlias = dict[str, cairo.SurfacePattern]

_ERRORS: set[str] = set()

@enum.unique
class Pen(enum.IntEnum):
    """
    Stroke pen id representing reMarkable tablet tools.

    Tool examples: ballpoint, fineliner, highlighter or eraser.
    """
    BALLPOINT_1 = 2
    BALLPOINT_2 = 15
    CALIGRAPHY = 21
    ERASER = 6
    ERASER_AREA = 8
    FINELINER_1 = 4
    FINELINER_2 = 17
    HIGHLIGHTER_1 = 5
    HIGHLIGHTER_2 = 18
    MARKER_1 = 3
    MARKER_2 = 16
    MECHANICAL_PENCIL_1 = 7
    MECHANICAL_PENCIL_2 = 13
    PAINTBRUSH_1 = 0
    PAINTBRUSH_2 = 12
    PENCIL_1 = 1
    PENCIL_2 = 14

    @classmethod
    def is_highlighter(cls, value: int) -> bool:
        return value in (cls.HIGHLIGHTER_1, cls.HIGHLIGHTER_2)

@enum.unique
class ColorIndex(enum.IntEnum):
    """
    Color index value.
    """
    BLACK = 0
    GRAY = 1
    WHITE = 2

    YELLOW = 3
    GREEN = 4
    PINK = 5

    BLUE = 6
    RED = 7

    GRAY_OVERLAP = 8

    @classmethod
    def _missing_(cls, value: tp.Any) -> ColorIndex:
        _ERRORS.add('Unknown color index: {}'.format(value))
        return ColorIndex.BLACK

class Color(tp.NamedTuple):
    red: float
    green: float
    blue: float
    alpha: float

class Style(tp.NamedTuple):
    tool_line: tp.Callable[..., tp.Any]
    color: Color
    join: cairo.LineJoin
    cap: cairo.LineCap
    dash: tp.List[int] = []
    brush: tp.Optional[BrushPatterns] = None

class Page(tp.NamedTuple):
    number: int

class PageEnd(tp.NamedTuple):
    number: int

class Layer(tp.NamedTuple):
    number: int

class Stroke(tp.NamedTuple):
    number: int
    pen: Pen
    color: ColorIndex
    width: float
    segments: tp.List['Segment']
    text: tp.Optional[str] = None

class Segment(tp.NamedTuple):
    number: int
    x: float
    y: float
    speed: float
    direction: float
    width: float
    pressure: float

DocumentItem = tp.Union[Page, PageEnd, Layer, Stroke, Segment]
Items = tp.Iterable[DocumentItem]

class Context(tp.NamedTuple):
    """
    Remt drawing context.
    """
    cr_surface: cairo.PDFSurface
    cr_ctx: cairo.Context
    pdf_doc: tp.Any  # Poppler.Document
    style: tp.Dict[Pen, Style]
    page_number: tp.Iterator[int]

Point = tp.Tuple[float, float]
Points = tp.List[tp.Tuple[float, float]]

class LinePath(tp.NamedTuple):
    """
    Line calculated using stroke and segment data of reMarkable tablet.

    :var width: Width of line.
    :var pressure: Line pressure level.
    :var points: List of line points.
    """
    width: float
    pressure: int
    points: Points

@dtc.dataclass(frozen=True)
@total_ordering
class TextSelection:
    """
    Information required to select text in a document with a horizontal
    line.

    The text selection defines ordering, which is used to sort the
    selections and to determine if two text selections overlap.

    :var x1: Start of line (x-axis coordinate).
    :var x2: End of line (x-axis coordinate).
    :var y: Y-axis coordinate of the horizontal line.
    """
    x1: float
    x2: float
    y: float

    def merge(self, other: 'TextSelection') -> 'TextSelection':
        """
        Merge two text selection objects.
        """
        return TextSelection(
            min(self.x1, other.x1),
            max(self.x2, other.x2),
            (self.y + other.y) / 2,
        )

    def overlaps(self, other: 'TextSelection', tolerance: float) -> bool:
        """
        Check if two text selection objects overlap.

        NOTE: It is assumed that this object and the other objects are
        sorted.

        :param other: Other text selection object.
        :param tolerance: Vertical line tolerance.
        """
        return other.y - self.y <= tolerance and other.x1 <= self.x2

    def __lt__(self, other: 'TextSelection') -> bool:
        return TEXT_SELECTION_ORDER(self) < TEXT_SELECTION_ORDER(other)

    def __eq__(self, other: tp.Any) -> bool:
        return TEXT_SELECTION_ORDER(self) == TEXT_SELECTION_ORDER(other)

# vim: sw=4:et:ai



#
# remt - note-taking support tool for reMarkable tablet
#
# Copyright (C) 2018-2021 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Unit tests for methods of data classes representing reMarkable tablet
information.
"""

from ..data import Pen, TextSelection

def test_pen_higlighter() -> None:
    """
    Test checking if a pen is a highlighter.
    """
    assert Pen.is_highlighter(Pen.HIGHLIGHTER_1)
    assert Pen.is_highlighter(Pen.HIGHLIGHTER_2)
    assert not Pen.is_highlighter(Pen.FINELINER_1)

def test_text_selection_order() -> None:
    """
    Test sorting of text selection.
    """
    items = [
        TextSelection(50, 100, 15),
        TextSelection(10, 40, 15),
        TextSelection(10, 100, 25),
        TextSelection(10, 100, 5),
    ]
    result = sorted(items)

    expected = [
        TextSelection(10, 100, 5),
        TextSelection(10, 40, 15),
        TextSelection(50, 100, 15),
        TextSelection(10, 100, 25),
    ]

    assert expected == list(result)

def test_text_selection_overlap() -> None:
    """
    Test overlapping check of text selections.
    """
    ts1 = TextSelection(10, 100, 15)
    ts2 = TextSelection(101, 200, 20)
    ts3 = TextSelection(100, 200, 20)

    assert not ts1.overlaps(ts2, 10)
    assert ts1.overlaps(ts3, 10)
    assert ts2.overlaps(ts3, 10)
    assert not ts1.overlaps(ts3, 1)

def test_text_selection_merge() -> None:
    """
    Test merging f text selections.
    """
    ts1 = TextSelection(10, 100, 15)
    ts2 = TextSelection(101, 200, 25)

    assert TextSelection(10, 200, 20) == ts1.merge(ts2)

# vim: sw=4:et:ai



#
# remt - note-taking support tool for reMarkable tablet
#
# Copyright (C) 2018-2021 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Command line commands unit tests.
"""

import os.path
import typing as tp
from datetime import datetime

from remt import cmd as rcmd
from remt import meta as rmeta
from remt.error import ConfigError

import pytest
from unittest import mock

def test_ls_line() -> None:
    """
    Test creating `ls` command basic output line.
    """
    meta = tp.cast(rmeta.Meta, {'uuid': 'xyz'})
    result = rcmd.ls_line('a/b', meta)
    assert result == 'a/b'

@mock.patch.object(rcmd, 'datetime')
def test_ls_line_long(mock_dt: mock.MagicMock) -> None:
    """
    Test creating `ls` command long output line.
    """
    meta = {
        'pinned': True,
        'bookmarked': True,
        'type': 'CollectionType',
        'lastModified': '1526115458925',
    }
    mock_dt.fromtimestamp.return_value = datetime(2018, 5, 12, 9, 57, 38)
    result = rcmd.ls_line_long('a/b', tp.cast(rmeta.Meta, meta))
    assert result == 'db 2018-05-12 09:57:38 a/b'

def test_ls_filter_path() -> None:
    """
    Test `ls` command metadata filtering with parent path.
    """
    meta = tp.cast(
        dict[str, rmeta.Meta],
        {
            'a': {'uuid': 'xyz-a'},
            'a/b': {'uuid': 'xyz-ab'},
            'a/c': {'uuid': 'xyz-ac'},
            'b': {'uuid': 'xyz-b'},
            'c': {'uuid': 'xyz-c'}
        }
    )
    result = rcmd.ls_filter_path(meta, 'a')
    expected = tp.cast(
        dict[str, rmeta.Meta],
        {
            'a/b': {'uuid': 'xyz-ab'},
            'a/c': {'uuid': 'xyz-ac'},
        }
    )
    assert result == expected

def test_ls_filter_parent_uuid() -> None:
    """
    Test `ls` command metadata filtering for items with parent identified
    by UUID.
    """
    meta = tp.cast(
        dict[str, rmeta.Meta],
        {
            'a': {'uuid': 'uuid-1'},
            'a/b': {'uuid': 'uuid-2', 'parent': 'uuid-1'},
            'a/c': {'uuid': 'uuid-3', 'parent': 'uuid-1'},
            'd': {'uuid': 'uuid-4'},
        }
    )
    result = rcmd.ls_filter_parent_uuid(meta, 'uuid-1')
    expected = tp.cast(
        dict[str, rmeta.Meta],
        {
            'a/b': {'uuid': 'uuid-2', 'parent': 'uuid-1'},
            'a/c': {'uuid': 'uuid-3', 'parent': 'uuid-1'},
        }
    )
    assert result == expected

def test_ls_filter_parent_uuid_null() -> None:
    """
    Test `ls` command metadata filtering for items with parent identified
    by UUID when UUID is null.
    """
    meta = tp.cast(
        dict[str, rmeta.Meta],
        {
            'a': {'uuid': 'uuid-1'},
            'a/b': {'uuid': 'uuid-2', 'parent': 'uuid-1'},
            'a/c': {'uuid': 'uuid-3', 'parent': 'uuid-1'},
            'd': {'uuid': 'uuid-4'},
        }
    )
    result = rcmd.ls_filter_parent_uuid(meta, None)

    # only items with no parents expected
    expected = tp.cast(
        dict[str, rmeta.Meta],
        {
            'a': {'uuid': 'uuid-1'},
            'd': {'uuid': 'uuid-4'},
        }
    )
    assert expected == result

def test_read_config_error() -> None:
    """
    Test if error is raised when no Remt configuration project is found.
    """
    with mock.patch.object(os.path, 'exists') as exists:
        exists.return_value = False

        with pytest.raises(ConfigError):
            rcmd.read_config()

def test_norm_path() -> None:
    """
    Test path normalisation.
    """
    result = rcmd.norm_path('a//b/c//d///')
    assert 'a/b/c/d' == result

def test_norm_path_leading_slashes() -> None:
    """
    Test path normalisation with leading slashes.
    """
    result = rcmd.norm_path('///a//b/c//d///')
    assert 'a/b/c/d' == result

# vim: sw=4:et:ai



#
# remt - note-taking support tool for reMarkable tablet
#
# Copyright (C) 2018-2021 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Tests for reading and parsing reMarkable tablet metadata.
"""

import os.path
import typing as tp

from remt import meta as rmeta
from remt.error import FileError, ConnectionError

import pytest
from unittest import mock

def test_fn_path() -> None:
    """
    Test creating UUID based path from metadata.
    """
    meta = tp.cast(rmeta.Meta, {'uuid': 'xyz'})
    result = rmeta.fn_path(meta, base='/x/y', ext='.met')
    assert '/x/y/xyz.met' == result

def test_read_meta() -> None:
    """
    Test reading reMarkable tablet metadata.
    """
    dirmeta = 'dir'

    with mock.patch('glob.glob') as mock_glob, \
            mock.patch('json.load') as mock_json_load, \
            mock.patch('builtins.open') as mock_open:  # noqa: F841

        mock_glob.side_effect = [
            # let's make read_meta resistant to some filesystem
            # inconsistencies by introducing non-matching files
            [
                'f1.content', 'f2.content', 'f3.content', 'f4.content',
                'f5.content', 'f6.content', 'f7.content',
            ],
            [
                'f1.metadata', 'f2.metadata', 'f3.metadata', 'f4.metadata',
                'f5.metadata', 'f6.metadata', 'f7.metadata',
            ],
        ]
        mock_json_load.side_effect = [
            {'visibleName': 'f1', 'deleted': False},
            {'pages': 3},

            # deleted filename shall not be visible in the results
            {'visibleName': 'f2', 'deleted': True},
            {'pages': 4},

            {'visibleName': 'f3'},
            {'pages': 5},

            {'visibleName': 'f4', 'parent': 'f3'},
            {'pages': 6},

            # in trash, shall not be visible in the results
            {'visibleName': 'f5', 'deleted': False, 'parent': 'trash'},
            {'pages': 9},

            # there is no parent 'has-parent', so it should be ignored
            {'visibleName': 'f6', 'deleted': False, 'parent': 'has-parent'},
            {'pages': 1},

            {'visibleName': 'f7'},
            {'pages': 12},
        ]
        result = rmeta.read_meta(dirmeta)
        assert ['f1', 'f3', 'f3/f4', 'f7'] == list(result)
        assert {'pages': 3} == result['f1']['content']  # type: ignore
        assert {'pages': 5} == result['f3']['content']  # type: ignore
        assert {'pages': 6} == result['f3/f4']['content']  # type: ignore
        assert {'pages': 12} == result['f7']['content']  # type: ignore

def test_fn_metadata_error() -> None:
    """
    Test if error is raised if there is no metadata for a path.
    """
    meta = {'a/b/': tp.cast(rmeta.Meta, {'uuid': 'xyz'})}

    with pytest.raises(FileError):
        rmeta.fn_metadata(meta, 'x/y')

def test_is_meta() -> None:
    """
    Test if a filename is reMarkable tablet meta file.
    """
    assert rmeta.is_meta('x.content')
    assert rmeta.is_meta('x.metadata')
    assert not rmeta.is_meta('x.pdf')

def test_find_meta_update() -> None:
    """
    Test finding reMarkable tablet meta files, which need to be updated in
    local cache.
    """
    files_remote = {
        'a': 10.0,
        'b': 10.0,
        'c': 10.0,
        'x': 11.0,  # updated meta file
        'y': 11.0,
        'n': 10.0,  # new meta file
    }
    files_local = {
        'a': 10.0,
        'b': 10.0,
        'c': 10.0,
        'x': 9.0,
        'y': 12.0,
        'm': 9.0,
    }

    result = rmeta.find_meta_update(files_remote, files_local)
    expected = ['n', 'x']
    assert expected == list(sorted(result))

def test_find_meta_delete() -> None:
    """
    Test finding reMarkable tablet meta files, which need to be removed
    from local cache.
    """
    files_remote = {
        'c': 10.0,
    }
    files_local = {
        'a': 10.0,  # to be removed
        'b': 10.0,  # to be removed
        'c': 10.0,
    }

    result = rmeta.find_meta_delete(files_remote, files_local)
    expected = ['a', 'b']
    assert expected == list(sorted(result))

def test_cache_dir() -> None:
    """
    Test generating cache directory for reMarkable tablet metadata.
    """
    with mock.patch.object(rmeta, 'getmac') as mock_getmac_mod, \
            mock.patch.object(os.path, 'expanduser') as mock_expand:

        mock_getmac_mod.get_mac_address.return_value = 'a:b:c'
        mock_expand.return_value = '/home/t-user'
        result = rmeta.cache_dir('a-host')
        assert '/home/t-user/.cache/remt/a:b:c' == result

def test_cache_dir_no_mac() -> None:
    """
    Test if error is raised when no MAC address is returned when generating
    cache directory for reMarkable tablet metadata.
    """
    with mock.patch.object(rmeta, 'getmac') as mock_getmac_mod:
        mock_getmac_mod.get_mac_address.return_value = None
        with pytest.raises(ConnectionError):
            rmeta.cache_dir('a-host')

def test_ls_local() -> None:
    """
    Test listing files in local cache.
    """
    with mock.patch('glob.glob') as mock_glob, \
            mock.patch('os.stat') as mock_stat:

        def create_stat(ts: float) -> mock.MagicMock:
            s = mock.MagicMock()
            s.st_mtime = ts
            return s

        mock_stat.side_effect = [create_stat(v) for v in range(1, 5)]
        mock_glob.return_value = [
            'x/y/1.content',
            'x/y/1.metadata',
            'x/y/2.content',
            'x/y/2.metadata',
        ]

        result = rmeta.ls_local('a-dir')

        expected = {
            '1.content': 1,
            '1.metadata': 2,
            '2.content': 3,
            '2.metadata': 4,
        }
        assert expected == result

@pytest.mark.asyncio
async def test_ls_remote() -> None:
    """
    Test listing metadata files from reMarkable tablet directory.
    """
    def create_fn(ts: float) -> mock.MagicMock:
        fn = mock.MagicMock()
        fn.filename = '{}.metadata'.format(ts)
        fn.attrs.mtime = ts
        return fn

    sftp = mock.MagicMock()
    sftp.readdir = mock.AsyncMock()
    sftp.readdir.return_value = [create_fn(v) for v in range(1, 5)]

    result = await rmeta.ls_remote(sftp)
    expected = {
        '1.metadata': 1,
        '2.metadata': 2,
        '3.metadata': 3,
        '4.metadata': 4,
    }
    assert expected == result

# vim: sw=4:et:ai



#
# remt - note-taking support tool for reMarkable tablet
#
# Copyright (C) 2018-2021 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Drawing tool calculations.

Drawing tools characteristics is as follows

+-------------------+----------+------+-------------------+
| Tool              | Pressure | Tilt |      Brush        |
+===================+==========+======+===================+
| Ballpoint         |    Y     |  N   |        Y          |
+-------------------+----------+------+-------------------+
| Fineliner         |    N     |  N   |        N          |
+-------------------+----------+------+-------------------+
| Marker            |    N     |  Y   |      marker       |
+-------------------+----------+------+-------------------+
| Pencil            |    Y     |  Y   |      pencil       |
+-------------------+----------+------+-------------------+
| Mechanical pencil |   Y (1)  |  N   | pencil-mechanical |
+-------------------+----------+------+-------------------+
| Paintbrush        |    Y     |  Y   |        N          |
+-------------------+----------+------+-------------------+
| Highlighter       |    N     |  N   |        N          |
+-------------------+----------+------+-------------------+
| Caligraphy        |    N     |  Y   |        N          |
+-------------------+----------+------+-------------------+
| Eraser            |    N     |  N   |        N          |
+-------------------+----------+------+-------------------+
| Erase Area        |    N     |  N   |        N (3)      |
+-------------------+----------+------+-------------------+

1. Mechanical pencil uses pressure to distinguish between two brush
   versions - lighter or darker.
2. Highlighter has static width of 30px.
3. Some shapes have an irregular edge, could it be due to a brush?

Use transparent color for highlighter tool only. All other tools should use
appropriate brushes at full opacity. For example, drawing with pencil in
exactly the same place does not make it darker. This allows to draw a
single stroke of varying width with multiple lines in Cairo, which does not
support single line of varying width.

Certain formulas are adapted from

   https://github.com/Wacom-Developer/sdk-for-ink-android/

"""

import cytoolz.itertoolz as itz  # type: ignore
import cytoolz.functoolz as ftz  # type: ignore
import math
import operator
import typing as tp
from collections.abc import Iterator
from functools import partial
from itertools import groupby

from . import data as rdata

T = tp.TypeVar('T')
V = tp.TypeVar('V')

PI_2 = math.pi / 2

LineConstWidthCalc = tp.Callable[[rdata.Stroke], float]
LineVarWidthCalc = tp.Callable[[rdata.Stroke, rdata.Segment], float]

def line_const(
        calc: LineConstWidthCalc, stroke: rdata.Stroke
) -> Iterator[rdata.LinePath]:
    """
    Return collection of lines of constant width and pressure between
    segments.

    :param calc: Width calculator.
    :param stroke: Stroke data to convert to line.
    """
    width = calc(stroke)
    segments = stroke.segments
    to_point = tp.cast(
        tp.Callable[[tp.Tuple[float, ...]], rdata.Point],
        operator.attrgetter('x', 'y')
    )

    points = [to_point(s) for s in segments]

    # set pressure to 1 as line is pressure insensitive
    yield rdata.LinePath(width, 1, points)

def line_var_pressure(
        calc: LineConstWidthCalc, stroke: rdata.Stroke
) -> Iterator[rdata.LinePath]:
    """
    Return collection of lines of constant width, but varying pressure
    between segments.

    :param calc: Width calculator.
    :param stroke: Stroke data to convert to line.
    """
    width = calc(stroke)
    segments = stroke.segments
    create_line = partial(rdata.LinePath, width)
    key = tp.cast(
        tp.Callable[[tp.Tuple[float, ...]], int],
        operator.itemgetter(2)
    )

    # extract positions and optimize by pressure level
    items = ((s.x, s.y, pressure_level(s)) for s in segments)
    by_pressure = group_line_segments(items, key)

    lines = (create_line(p, pt) for p, pt in by_pressure)
    yield from lines

def line_var_width_pressure(
        calc: LineVarWidthCalc, stroke: rdata.Stroke
) -> Iterator[rdata.LinePath]:
    """
    Return collection of lines having varying width or varying pressure
    between segments.

    :param calc: Width calculator.
    :param stroke: Stroke data to convert to line.
    """
    segments = stroke.segments

    # convert width to key; by rounding at 3rd decimal place we shall still
    # get more than 1200 dpi out of it
    key_width = ftz.compose(
        int,
        partial(operator.mul, 1000),
        partial(calc, stroke),
    )

    # extract width and pressure
    key = tp.cast(
        tp.Callable[[tp.Tuple[float, ...]], tp.Tuple[int, int]],
        operator.itemgetter(2, 3)
    )

    items = ((s.x, s.y, key_width(s), pressure_level(s)) for s in segments)
    groups = group_line_segments(items, key)
    lines = (rdata.LinePath(w / 1000, p, pt) for (w, p), pt in groups)
    yield from lines

def group_line_segments(
        segments: Iterator[T], key: tp.Callable[[T], V]
) -> Iterator[tp.Tuple[V, rdata.Points]]:
    """
    Group line segments using key function.
    """
    to_points = ftz.compose(list, partial(itz.pluck, [0, 1]))

    items = groupby(segments, key)
    by_key = [(k, to_points(v)) for k, v in items]
    if not by_key:
        return

    # yield each segment; add first point of next segment to the current
    # segment
    pairs = itz.sliding_window(2, by_key)
    result = ((k, pt + pt_n[:1]) for (k, pt), (_, pt_n) in pairs)
    yield from result

    # if last segment is not just a point, then return it as well
    last = by_key[-1]
    if len(last[1]) > 1:
        yield last

def pressure_level(segment: rdata.Segment) -> int:
    """
    Calculate pressure level for a stroke segment.
    """
    return min(99, int(segment.pressure * 100))

def calc_width_fineliner(stroke: rdata.Stroke) -> float:
    """
    Calculate fineliner width.

    :param stroke: Stroke data.
    """
    return calc_width_standard(stroke)

def calc_width_marker(stroke: rdata.Stroke, segment: rdata.Segment) -> float:
    """
    Calculate marker tool width.
    """
    return calc_width_tilt(segment) + 0.5

def calc_width_pencil(stroke: rdata.Stroke, segment: rdata.Segment) -> float:
    """
    Calculate marker tool width.
    """
    return calc_width_tilt(segment) + 0.5

def calc_width_painbrush(stroke: rdata.Stroke, segment: rdata.Segment) \
        -> float:
    """
    Calculate marker tool width.
    """
    width = calc_width_tilt(segment) + 0.5
    return apply_pressure(segment, width)

def calc_width_mechanical_pencil(stroke: rdata.Stroke) -> float:
    """
    Calculate mechanical pencil width.

    :param stroke: Stroke data.
    """
    return 16 * stroke.width - 27

def calc_width_eraser(stroke: rdata.Stroke) -> float:
    """
    Calculate eraser width.

    :param stroke: Stroke data.
    """
    return 1280 * stroke.width ** 2 - 4800 * stroke.width + 4510

def calc_width_ballpoint(stroke: rdata.Stroke, segment: rdata.Segment) \
        -> float:
    """
    Calculate ballpoint tool line width.

    :param stroke: Stroke data.
    :param segment: Segment data.
    """
    width = calc_width_standard(stroke)
    return apply_pressure(segment, width)

def calc_width_caligraphy(stroke: rdata.Stroke, segment: rdata.Segment) \
        -> float:
    """
    Calculate caligraphy tool line width.

    :param stroke: Stroke data.
    :param segment: Segment data.
    """
    return calc_width_tilt(segment)

def calc_width_standard(stroke: rdata.Stroke) -> float:
    """
    Calculate width for a tool without tilt.

    :param stroke: Stroke data.
    """
    return 32 * stroke.width ** 2 - 116 * stroke.width + 107

def calc_width_tilt(segment: rdata.Segment) -> float:
    """
    Calculate width for a tool with tilt applied.

    :param segment: Segment data.
    """
    return segment.width / PI_2 * 1.5

def apply_pressure(segment: rdata.Segment, width: float) -> float:
    """
    Apply pressure to the width of a tool.

    :param segment: Segment data.
    :param width: Width to be changed by pressure.
    """
    return width + segment.pressure ** 2048

line_unknown = partial(line_const, lambda s: rdata.PEN_WIDTH_UKNOWN)
line_ballpoint = partial(line_var_width_pressure, calc_width_ballpoint)
line_fineliner = partial(line_const, calc_width_fineliner)
line_marker = partial(line_var_width_pressure, calc_width_marker)
line_pencil = partial(line_var_width_pressure, calc_width_pencil)
line_mechanical_pencil = partial(
    line_var_pressure, calc_width_mechanical_pencil
)
line_paintbrush = partial(line_var_width_pressure, calc_width_painbrush)
line_highlighter = partial(line_const, lambda s: rdata.PEN_WIDTH_HIGHLIGHTER)
line_caligraphy = partial(line_var_width_pressure, calc_width_caligraphy)
line_eraser = partial(line_const, calc_width_eraser)
line_erase_area = partial(line_const, lambda s: rdata.PEN_WIDTH_ERASE_AREA)

# vim: sw=4:et:ai



#
# remt - note-taking support tool for reMarkable tablet
#
# Copyright (C) 2018-2021 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Read and parse reMarkable tablet metadata.
"""

import getmac  # type: ignore
import glob
import json
import logging
import operator
import os
import typing as tp
from cytoolz.functoolz import compose_left  # type: ignore
from datetime import datetime
from functools import partial
from asyncssh.sftp import SFTPClient

from remt.error import FileError, ConnectionError

logger = logging.getLogger(__name__)

BASE_DIR = '/home/root/.local/share/remarkable/xochitl'
META_EXT = {'.content', '.metadata'}

fn_local = os.path.join

class MetaContent(tp.TypedDict):
    fileType: str
    lastOpenedPage: int
    lineHeight: int
    pageCount: int

class Meta(tp.TypedDict):
    uuid: str
    deleted: bool
    lastModified: str
    metadatamodified: bool
    modified: bool
    parent: str
    pinned: bool
    synced: bool
    type: str
    version: int
    visibleName: str
    content: MetaContent

MetaDir: tp.TypeAlias = dict[str, Meta]
MetaPredicate: tp.TypeAlias = tp.Callable[[Meta], bool]

def fn_path(data: Meta, base: str=BASE_DIR, ext: str='.*') -> str:
    """
    Having metadata object create UUID based path of a file created by
    reMarkable tablet.

    :param data: Metadata object.
    """
    return '{}/{}{}'.format(base, data['uuid'], ext)

def fn_remote(fn: str) -> str:
    """
    Return path to a file on reMarkable tablet.
    """
    return os.path.join(BASE_DIR, fn)

def create_metadata(is_dir: bool, parent_uuid: str, name: str) -> Meta:
    now = datetime.utcnow()
    tstamp = int(now.timestamp() * 1000)
    type = 'CollectionType' if is_dir else 'DocumentType'
    data = {
        'deleted': False,
        'lastModified': str(tstamp),
        'metadatamodified': True,
        'modified': True,
        'parent': parent_uuid,
        'pinned': False,
        'synced': False,
        'type': type,
        'version': 0,
        'visibleName': name,
    }
    return tp.cast(Meta, data)

def to_path(data: Meta, meta: MetaDir) -> str:
    parent = data.get('parent')
    name = data['visibleName']
    if parent:
        return to_path(meta[parent], meta) + '/' + name
    else:
        return name

def cleanup_meta(meta: MetaDir) -> MetaDir:
    """
    Cleanup reMarkable tablet metadata records.

    The cleanup is

    - update uuid values of metadata records
    - resolve file path of a metadata record
    - remove metadata records in trash or marked as deleted
    - remove metadata records, which parent is in trash or deleted
    """
    is_toplevel: MetaPredicate = lambda v: not v.get('parent')
    is_deleted: MetaPredicate = lambda v: \
        v.get('parent') == 'trash' or bool(v.get('deleted'))
    parent_in_meta: MetaPredicate = lambda v: v.get('parent') in meta

    # assign uuid and drop deleted items
    meta = {
        k: v | {'uuid': k} for k, v in meta.items()  # type: ignore
        if not is_deleted(v)
    }

    # keep top-level items and items which parent exists; drop the item
    # otherwise
    meta = {
        k: v for k, v in meta.items()
        if is_toplevel(v) or parent_in_meta(v)
    }
    return {to_path(data, meta): data for data in meta.values()}

def fn_metadata(meta: MetaDir, path: str) -> Meta:
    """
    Get reMarkable tablet file metadata or raise file not found error if no
    metadata for path is found.

    :param meta: reMarkable tablet metadata.
    :param path: Path of reMarkable tablet file.
    """
    data = meta.get(path)
    if not data:
        raise FileError('File or directory not found: {}'.format(path))
    return data

def is_meta(fn: str) -> bool:
    """
    Check if a filename is reMarkable tablet meta file.

    :param fn: Filename to be checked.
    """
    return any(fn.endswith(ext) for ext in META_EXT)

async def ls_remote(sftp: SFTPClient) -> dict[str, float]:
    """
    List reMarkable tablet meta files.
    """
    files = await sftp.readdir(BASE_DIR)
    meta_files = (f for f in files if is_meta(f.filename))  # type: ignore
    return {f.filename: f.attrs.mtime for f in meta_files}  # type: ignore

def ls_local(dir_meta: str) -> dict[str, float]:
    """
    List reMarkable tablet meta files in local cache.
    """
    files = glob.glob(fn_local(dir_meta, '*'))
    return {os.path.basename(fn): os.stat(fn).st_mtime for fn in files}

def find_meta_update(
        files_remote: dict[str, float], files_local: dict[str, float]
    ) -> tp.Iterable[str]:
    """
    Find reMarkable tablet meta files, which need to be updated in local
    cache.
    """
    items = (f for f, t in files_remote.items() if t > files_local.get(f, 0))
    yield from items

def find_meta_delete(
        files_remote: dict[str, float], files_local: dict[str, float],
    ) -> tp.Iterable[str]:
    """
    Find reMarkable tablet meta files, which need to be removed from local
    cache.
    """
    yield from (f for f in files_local if f not in files_remote)

async def sync_meta(sftp: SFTPClient, dir_meta: str) -> None:
    """
    Synchronize metadata from reMarkable tablet to local cache.

    :param sftp: SFTP connection object.
    :param dir_meta: Local cache directory with reMarkable tablet metadata.
    """
    files_remote = await ls_remote(sftp)
    files_local = ls_local(dir_meta)

    # update new and out-of-date meta files
    files = find_meta_update(files_remote, files_local)
    files = [fn_remote(f) for f in files]
    await sftp.mget(files, dir_meta, preserve=True, recurse=True)
    logger.info('fetched {} files into local cache'.format(len(files)))

    # remove metadata, which does not exist on reMarkable tablet anymore
    files = list(find_meta_delete(files_remote, files_local))
    for fn in files:
        assert fn.endswith('.content') or fn.endswith('.metadata')
        os.unlink(fn_local(dir_meta, fn))
    logger.info('removed {} files from local cache'.format(len(files)))

def read_meta(dir_meta: str) -> MetaDir:
    """
    Read reMarkable tablet metadata.
    """
    to_uuid = compose_left(
        os.path.basename,
        os.path.splitext,
        operator.itemgetter(0),
    )
    load_json = compose_left(open, json.load)
    list_files = compose_left(partial(fn_local, dir_meta), glob.glob)

    files_content = list_files('*.content')
    files_content = {to_uuid(fn): fn for fn in files_content}

    files = list_files('*.metadata')
    files = ((fn, to_uuid(fn)) for fn in files)
    files = ((fn, u) for fn, u in files if u in files_content)

    # load metadata file and content file
    data = (
        (u, load_json(fn), load_json(files_content[u]))
        for fn, u in files
    )
    meta = {u: m | {'content': c} for u, m, c in data}
    return cleanup_meta(meta)

def cache_dir(host: str) -> str:
    """
    Generate directory for reMarkable tablet metadata cache.

    :param host: reMarkable tablet hostname.
    """
    mac = getmac.get_mac_address(hostname=host)
    if mac is None:
        raise ConnectionError(
            'Cannot get MAC address for reMarkable tablet at {}'.format(host)
        )
    base = os.path.expanduser('~')
    return os.path.join(base, '.cache/remt', mac)

# vim: sw=4:et:ai



#
# remt - note-taking support tool for reMarkable tablet
#
# Copyright (C) 2018-2021 by Artur Wroblewski <wrobell@riseup.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
reMarkable strokes drawing using Cairo library.
"""

import cairo
import cytoolz.itertoolz as itz  # type: ignore
import io
import itertools
import logging
import os.path
import pkgutil
import typing as tp
from collections import defaultdict
from contextlib import contextmanager
from functools import singledispatch, partial
from itertools import accumulate

from . import const, tool
from . import data as rdata
from .pdf import pdf_scale, Poppler

logger = logging.getLogger(__name__)

ALPHA = 0.1

COLOR_STROKE = {
    rdata.ColorIndex.BLACK: rdata.Color(0, 0, 0, 1),
    rdata.ColorIndex.GRAY: rdata.Color(0.5, 0.5, 0.5, 1),
    rdata.ColorIndex.WHITE: rdata.Color(1, 1, 1, 1),
    rdata.ColorIndex.BLUE: rdata.Color(0, 0.38, 0.79, 1),
    rdata.ColorIndex.RED: rdata.Color(0.85, 0.02, 0.2, 1),
    rdata.ColorIndex.YELLOW: rdata.Color(1, 0.81, 0.0, ALPHA),
    rdata.ColorIndex.GREEN: rdata.Color(0.66, 0.98, 0.36, ALPHA),
    rdata.ColorIndex.PINK: rdata.Color(1, 0.33, 0.81, ALPHA),
    rdata.ColorIndex.GRAY_OVERLAP: rdata.Color(0.5, 0.5, 0.5, ALPHA),
}

STYLE_DEFAULT = rdata.Style(
    tool.line_const,
    COLOR_STROKE[rdata.ColorIndex.BLACK],
    cairo.LINE_JOIN_ROUND,
    cairo.LINE_CAP_ROUND,
)
style_default = STYLE_DEFAULT._replace

STYLE_HIGHLIGHTER = rdata.Style(
    tool.line_highlighter,
    COLOR_STROKE[rdata.ColorIndex.YELLOW],
    cairo.LINE_JOIN_ROUND,
    cairo.LINE_CAP_BUTT,
)

STYLE_ERASER = rdata.Style(
    tool.line_eraser,
    COLOR_STROKE[rdata.ColorIndex.WHITE],
    cairo.LINE_JOIN_ROUND,
    cairo.LINE_CAP_ROUND,
)

STYLE_UNKNOWN_TOOL = style_default(tool_line=tool.line_unknown, dash=[1, 5])

DEFAULT_COLOR = COLOR_STROKE[rdata.ColorIndex.BLACK]

path_brush = partial(os.path.join, 'brush')

@singledispatch
def draw(item: tp.Any, context: rdata.Context) -> None:
    raise NotImplementedError('Unknown item to draw: {}'.format(item))

@draw.register
def _page(page: rdata.Page, context: rdata.Context) -> None:
    surface = context.cr_surface
    page_number = next(context.page_number)
    if page_number:
        surface.show_page()

    if context.pdf_doc:
        # get page and set size of the current page of the cairo surface
        pdf_page = context.pdf_doc.get_page(page_number)
        w, h = pdf_page.get_size()
        surface.set_size(w, h)

        cr = context.cr_ctx
        # render for printing to keep the quality of the document
        pdf_page.render_for_printing(cr)

        # render remarkable lines data at scale to fit the document
        cr.save()  # to be restored at page end
        factor = pdf_scale(pdf_page)
        cr.scale(factor, factor)

@draw.register
def _page_end(page: rdata.PageEnd, context: rdata.Context) -> None:
    if context.pdf_doc:
        context.cr_ctx.restore()

@draw.register
def _layer(layer: rdata.Layer, context: rdata.Context) -> None:
    pass

@draw.register
def _stroke(stroke: rdata.Stroke, context: rdata.Context) -> None:
    style = context.style[stroke.pen]
    if style is STYLE_UNKNOWN_TOOL:
        logger.debug('Not supported pen for stroke: {}'.format(stroke))

    if stroke.color not in COLOR_STROKE:
        rdata._ERRORS.add('No color defined for index: {}'.format(stroke.color))

    if stroke.pen in (rdata.Pen.ERASER_AREA, rdata.Pen.ERASER):
        color = style.color
    else:
        color = COLOR_STROKE.get(stroke.color, DEFAULT_COLOR)

    cr = context.cr_ctx
    cr.save()

    draw_stroke = partial(draw_fill, cr) \
        if stroke.pen == rdata.Pen.ERASER_AREA else cr.stroke

    cr.set_source_rgba(*color)
    cr.set_line_join(style.join)
    cr.set_line_cap(style.cap)

    if style.dash:
        cr.set_dash(style.dash, 0)

    lines = style.tool_line(stroke)
    for line in lines:
        cr.set_line_width(line.width)

        if style.brush:
            brush = style.brush[line.pressure]
            cr.set_source(brush)

        draw_line(line, cr, draw_stroke)

    cr.restore()

def draw_line(
        line: rdata.LinePath,
        cr: cairo.Context,
        draw_stroke: tp.Any
    ) -> None:
    """
    Draw line using Cairo context.

    :param line: Line to draw.
    :param cr: Cairo context.
    :param draw_stroke: Drawing stroke function (i.e. line or filled area).
    """
    # on new path, the position of point is undefined and first `line_to`
    # call acts as `move_to`
    cr.new_path()
    for x, y in line.points:
        cr.line_to(x, y)
    draw_stroke()

def draw_fill(cr: cairo.Context) -> None:
    """
    Draw Cairo shape and fill.
    """
    cr.close_path()
    cr.fill()

@contextmanager
def draw_context(
        pdf_doc: tp.Optional[Poppler.Document], fn_out: str
) -> tp.Generator[rdata.Context, None, None]:
    """
    Create Remt project drawing context.
    """
    # mechanical_pencil = style_default(
    #     tool_line=tool.line_mechanical_pencil,
    #     brush=load_brush_dir('mechanical-pencil'),
    # )
    #
    # marker = style_default(
    #     tool_line=tool.line_marker,
    #     brush=load_brush_dir('marker'),
    # )
    #
    # pencil = style_default(
    #     tool_line=tool.line_pencil,
    #     brush=load_brush_dir('pencil'),
    # )

    styles = defaultdict(lambda: STYLE_UNKNOWN_TOOL, {
        rdata.Pen.BALLPOINT_1: style_default(tool_line=tool.line_ballpoint),
        rdata.Pen.BALLPOINT_2: style_default(tool_line=tool.line_ballpoint),

        rdata.Pen.FINELINER_1: style_default(tool_line=tool.line_fineliner),
        rdata.Pen.FINELINER_2: style_default(tool_line=tool.line_fineliner),

        rdata.Pen.PAINTBRUSH_1: style_default(tool_line=tool.line_paintbrush),
        rdata.Pen.PAINTBRUSH_2: style_default(tool_line=tool.line_paintbrush),

        rdata.Pen.MARKER_1: style_default(tool_line=tool.line_marker),
        rdata.Pen.MARKER_2: style_default(tool_line=tool.line_marker),

        rdata.Pen.HIGHLIGHTER_1: STYLE_HIGHLIGHTER,
        rdata.Pen.HIGHLIGHTER_2: STYLE_HIGHLIGHTER,

        rdata.Pen.CALIGRAPHY: style_default(tool_line=tool.line_caligraphy),

        # rdata.Pen.PENCIL_1: pencil,
        # rdata.Pen.PENCIL_2: pencil,
        # rdata.Pen.MECHANICAL_PENCIL_1: mechanical_pencil,
        # rdata.Pen.MECHANICAL_PENCIL_2: mechanical_pencil,

        rdata.Pen.ERASER: STYLE_ERASER,
        rdata.Pen.ERASER_AREA: STYLE_ERASER._replace(
            tool_line=tool.line_erase_area
        ),
    })

    surface = cairo.PDFSurface(fn_out, const.PAGE_WIDTH, const.PAGE_HEIGHT)
    try:
        cr_ctx = cairo.Context(surface)
        yield rdata.Context(
            surface,
            cr_ctx,
            pdf_doc,
            styles,
            itertools.count(),
        )
    finally:
        surface.finish()

# def load_brush_dir(fn: str) -> rdata.BrushPatterns:
#     replace_null = lambda prev, v: v if v[1] else (v[0], prev[1])
#
#     brushes = ((i, load_brush(fn, i)) for i in range(100))
#     items = accumulate(brushes, replace_null, initial=(None, None))
#     items = itz.drop(1, items)
#     result = tp.cast(rdata.BrushPatterns, dict(items))
#     assert list(result) == list(range(100))
#     if any(v is None for v in result.values()):
#         raise ValueError('Invalid data for brush: {}'.format(fn))
#     return result

def load_brush(fn: str, n: int) -> tp.Optional[cairo.SurfacePattern]:
    try:
        fn_brush = path_brush(fn, '{:02d}.png'.format(n))
        data = pkgutil.get_data('remt', fn_brush)
    except FileNotFoundError:
        return None
    else:
        assert data is not None
        img = cairo.ImageSurface.create_from_png(io.BytesIO(data))
        brush = cairo.SurfacePattern(img)
        brush.set_extend(cairo.EXTEND_REPEAT)
        return brush

# vim: sw=4:et:ai
